import pandas as pd

from yfinance.data import TickerData

class Holders:
    _SCRAPE_URL_ = 'https://finance.yahoo.com/quote'

    def __init__(self, data: TickerData, proxy=None):
        self._data = data
        self.proxy = proxy

        self._major = None
        self._institutional = None
        self._mutualfund = None

    @property
    def major(self) -> pd.DataFrame:
        if self._major is None:
            self._scrape(self.proxy)
        return self._major

    @property
    def institutional(self) -> pd.DataFrame:
        if self._institutional is None:
            self._scrape(self.proxy)
        return self._institutional

    @property
    def mutualfund(self) -> pd.DataFrame:
        if self._mutualfund is None:
            self._scrape(self.proxy)
        return self._mutualfund

    def _scrape(self, proxy):
        ticker_url = "{}/{}".format(self._SCRAPE_URL_, self._data.ticker)
        try:
            resp = self._data.cache_get(ticker_url + '/holders', proxy)
            holders = pd.read_html(resp.text)
        except Exception:
            holders = []

        if len(holders) >= 3:
            self._major = holders[0]
            self._institutional = holders[1]
            self._mutualfund = holders[2]
        elif len(holders) >= 2:
            self._major = holders[0]
            self._institutional = holders[1]
        elif len(holders) >= 1:
            self._major = holders[0]

        if self._institutional is not None:
            if 'Date Reported' in self._institutional:
                self._institutional['Date Reported'] = pd.to_datetime(
                    self._institutional['Date Reported'])
            if '% Out' in self._institutional:
                self._institutional['% Out'] = self._institutional[
                                                   '% Out'].str.replace('%', '').astype(float) / 100

        if self._mutualfund is not None:
            if 'Date Reported' in self._mutualfund:
                self._mutualfund['Date Reported'] = pd.to_datetime(
                    self._mutualfund['Date Reported'])
            if '% Out' in self._mutualfund:
                self._mutualfund['% Out'] = self._mutualfund[
                                                '% Out'].str.replace('%', '').astype(float) / 100



import datetime
import json

import pandas as pd
import numpy as np

from yfinance import utils
from yfinance.data import TickerData
from yfinance.exceptions import YFinanceDataException, YFinanceException


class Fundamentals:

    def __init__(self, data: TickerData, proxy=None):
        self._data = data
        self.proxy = proxy

        self._earnings = None
        self._financials = None
        self._shares = None

        self._financials_data = None
        self._fin_data_quote = None
        self._basics_already_scraped = False
        self._financials = Financials(data)

    @property
    def financials(self) -> "Financials":
        return self._financials

    @property
    def earnings(self) -> dict:
        if self._earnings is None:
            self._scrape_earnings(self.proxy)
        return self._earnings

    @property
    def shares(self) -> pd.DataFrame:
        if self._shares is None:
            self._scrape_shares(self.proxy)
        return self._shares

    def _scrape_basics(self, proxy):
        if self._basics_already_scraped:
            return
        self._basics_already_scraped = True

        self._financials_data = self._data.get_json_data_stores('financials', proxy)
        try:
            self._fin_data_quote = self._financials_data['QuoteSummaryStore']
        except KeyError:
            err_msg = "No financials data found, symbol may be delisted"
            print('- %s: %s' % (self._data.ticker, err_msg))
            return None

    def _scrape_earnings(self, proxy):
        self._scrape_basics(proxy)
        # earnings
        self._earnings = {"yearly": pd.DataFrame(), "quarterly": pd.DataFrame()}
        if self._fin_data_quote is None:
            return
        if isinstance(self._fin_data_quote.get('earnings'), dict):
            try:
                earnings = self._fin_data_quote['earnings']['financialsChart']
                earnings['financialCurrency'] = self._fin_data_quote['earnings'].get('financialCurrency', 'USD')
                self._earnings['financialCurrency'] = earnings['financialCurrency']
                df = pd.DataFrame(earnings['yearly']).set_index('date')
                df.columns = utils.camel2title(df.columns)
                df.index.name = 'Year'
                self._earnings['yearly'] = df

                df = pd.DataFrame(earnings['quarterly']).set_index('date')
                df.columns = utils.camel2title(df.columns)
                df.index.name = 'Quarter'
                self._earnings['quarterly'] = df
            except Exception:
                pass

    def _scrape_shares(self, proxy):
        self._scrape_basics(proxy)
        # shares outstanding
        try:
            # keep only years with non None data
            available_shares = [shares_data for shares_data in
                                self._financials_data['QuoteTimeSeriesStore']['timeSeries']['annualBasicAverageShares']
                                if
                                shares_data]
            shares = pd.DataFrame(available_shares)
            shares['Year'] = shares['asOfDate'].agg(lambda x: int(x[:4]))
            shares.set_index('Year', inplace=True)
            shares.drop(columns=['dataId', 'asOfDate',
                                 'periodType', 'currencyCode'], inplace=True)
            shares.rename(
                columns={'reportedValue': "BasicShares"}, inplace=True)
            self._shares = shares
        except Exception:
            pass


class Financials:
    def __init__(self, data: TickerData):
        self._data = data
        self._income_time_series = {}
        self._balance_sheet_time_series = {}
        self._cash_flow_time_series = {}
        self._income_scraped = {}
        self._balance_sheet_scraped = {}
        self._cash_flow_scraped = {}

    def get_income_time_series(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._income_time_series
        if freq not in res:
            res[freq] = self._fetch_time_series("income", freq, proxy=None)
        return res[freq]

    def get_balance_sheet_time_series(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._balance_sheet_time_series
        if freq not in res:
            res[freq] = self._fetch_time_series("balance-sheet", freq, proxy=None)
        return res[freq]

    def get_cash_flow_time_series(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._cash_flow_time_series
        if freq not in res:
            res[freq] = self._fetch_time_series("cash-flow", freq, proxy=None)
        return res[freq]

    def _fetch_time_series(self, name, timescale, proxy=None):
        # Fetching time series preferred over scraping 'QuoteSummaryStore',
        # because it matches what Yahoo shows. But for some tickers returns nothing, 
        # despite 'QuoteSummaryStore' containing valid data.

        allowed_names = ["income", "balance-sheet", "cash-flow"]
        allowed_timescales = ["yearly", "quarterly"]

        if name not in allowed_names:
            raise ValueError("Illegal argument: name must be one of: {}".format(allowed_names))
        if timescale not in allowed_timescales:
            raise ValueError("Illegal argument: timescale must be one of: {}".format(allowed_names))

        try:
            statement = self._create_financials_table(name, timescale, proxy)

            if statement is not None:
                return statement
        except YFinanceException as e:
            print(f"- {self._data.ticker}: Failed to create {name} financials table for reason: {repr(e)}")
        return pd.DataFrame()

    def _create_financials_table(self, name, timescale, proxy):
        if name == "income":
            # Yahoo stores the 'income' table internally under 'financials' key
            name = "financials"

        keys = self._get_datastore_keys(name, proxy)
        try:
            return self.get_financials_time_series(timescale, keys, proxy)
        except Exception as e:
            pass

    def _get_datastore_keys(self, sub_page, proxy) -> list:
        data_stores = self._data.get_json_data_stores(sub_page, proxy)

        # Step 1: get the keys:
        def _finditem1(key, obj):
            values = []
            if isinstance(obj, dict):
                if key in obj.keys():
                    values.append(obj[key])
                for k, v in obj.items():
                    values += _finditem1(key, v)
            elif isinstance(obj, list):
                for v in obj:
                    values += _finditem1(key, v)
            return values

        try:
            keys = _finditem1("key", data_stores['FinancialTemplateStore'])
        except KeyError as e:
            raise YFinanceDataException("Parsing FinancialTemplateStore failed, reason: {}".format(repr(e)))

        if not keys:
            raise YFinanceDataException("No keys in FinancialTemplateStore")
        return keys

    def get_financials_time_series(self, timescale, keys: list, proxy=None) -> pd.DataFrame:
        timescale_translation = {"yearly": "annual", "quarterly": "quarterly"}
        timescale = timescale_translation[timescale]

        # Step 2: construct url:
        ts_url_base = \
            "https://query2.finance.yahoo.com/ws/fundamentals-timeseries/v1/finance/timeseries/{0}?symbol={0}" \
                .format(self._data.ticker)

        url = ts_url_base + "&type=" + ",".join([timescale + k for k in keys])
        # Yahoo returns maximum 4 years or 5 quarters, regardless of start_dt:
        start_dt = datetime.datetime(2016, 12, 31)
        end = pd.Timestamp.utcnow().ceil("D")
        url += "&period1={}&period2={}".format(int(start_dt.timestamp()), int(end.timestamp()))

        # Step 3: fetch and reshape data
        json_str = self._data.cache_get(url=url, proxy=proxy).text
        json_data = json.loads(json_str)
        data_raw = json_data["timeseries"]["result"]
        # data_raw = [v for v in data_raw if len(v) > 1] # Discard keys with no data
        for d in data_raw:
            del d["meta"]

        # Now reshape data into a table:
        # Step 1: get columns and index:
        timestamps = set()
        data_unpacked = {}
        for x in data_raw:
            for k in x.keys():
                if k == "timestamp":
                    timestamps.update(x[k])
                else:
                    data_unpacked[k] = x[k]
        timestamps = sorted(list(timestamps))
        dates = pd.to_datetime(timestamps, unit="s")
        df = pd.DataFrame(columns=dates, index=list(data_unpacked.keys()))
        for k, v in data_unpacked.items():
            if df is None:
                df = pd.DataFrame(columns=dates, index=[k])
            df.loc[k] = {pd.Timestamp(x["asOfDate"]): x["reportedValue"]["raw"] for x in v}

        df.index = df.index.str.replace("^" + timescale, "", regex=True)

        # Reorder table to match order on Yahoo website
        df = df.reindex([k for k in keys if k in df.index])
        df = df[sorted(df.columns, reverse=True)]

        return df

    def get_income_scrape(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._income_scraped
        if freq not in res:
            res[freq] = self._scrape("income", freq, proxy=None)
        return res[freq]

    def get_balance_sheet_scrape(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._balance_sheet_scraped
        if freq not in res:
            res[freq] = self._scrape("balance-sheet", freq, proxy=None)
        return res[freq]

    def get_cash_flow_scrape(self, freq="yearly", proxy=None) -> pd.DataFrame:
        res = self._cash_flow_scraped
        if freq not in res:
            res[freq] = self._scrape("cash-flow", freq, proxy=None)
        return res[freq]

    def _scrape(self, name, timescale, proxy=None):
        # Backup in case _fetch_time_series() fails to return data

        allowed_names = ["income", "balance-sheet", "cash-flow"]
        allowed_timescales = ["yearly", "quarterly"]

        if name not in allowed_names:
            raise ValueError("Illegal argument: name must be one of: {}".format(allowed_names))
        if timescale not in allowed_timescales:
            raise ValueError("Illegal argument: timescale must be one of: {}".format(allowed_names))

        try:
            statement = self._create_financials_table_old(name, timescale, proxy)

            if statement is not None:
                return statement
        except YFinanceException as e:
            print(f"- {self._data.ticker}: Failed to create financials table for {name} reason: {repr(e)}")
        return pd.DataFrame()

    def _create_financials_table_old(self, name, timescale, proxy):
        data_stores = self._data.get_json_data_stores("financials", proxy)

        # Fetch raw data
        if not "QuoteSummaryStore" in data_stores:
            raise YFinanceDataException(f"Yahoo not returning legacy financials data")
        data = data_stores["QuoteSummaryStore"]

        if name == "cash-flow":
            key1 = "cashflowStatement"
            key2 = "cashflowStatements"
        elif name == "balance-sheet":
            key1 = "balanceSheet"
            key2 = "balanceSheetStatements"
        else:
            key1 = "incomeStatement"
            key2 = "incomeStatementHistory"
        key1 += "History"
        if timescale == "quarterly":
            key1 += "Quarterly"
        if key1 not in data or data[key1] is None or key2 not in data[key1]:
            raise YFinanceDataException(f"Yahoo not returning legacy {name} financials data")
        data = data[key1][key2]

        # Tabulate
        df = pd.DataFrame(data)
        if len(df) == 0:
            raise YFinanceDataException(f"Yahoo not returning legacy {name} financials data")
        df = df.drop(columns=['maxAge'])
        for col in df.columns:
            df[col] = df[col].replace('-', np.nan)
        df.set_index('endDate', inplace=True)
        try:
            df.index = pd.to_datetime(df.index, unit='s')
        except ValueError:
            df.index = pd.to_datetime(df.index)
        df = df.T
        df.columns.name = ''
        df.index.name = 'Breakdown'
        # rename incorrect yahoo key
        df.rename(index={'treasuryStock': 'gainsLossesNotAffectingRetainedEarnings'}, inplace=True)

        # Upper-case first letter, leave rest unchanged:
        s0 = df.index[0]
        df.index = [s[0].upper()+s[1:] for s in df.index]

        return df



import functools
from functools import lru_cache

import hashlib
from base64 import b64decode
usePycryptodome = False  # slightly faster
# usePycryptodome = True
if usePycryptodome:
    from Crypto.Cipher import AES
    from Crypto.Util.Padding import unpad
else:
    from cryptography.hazmat.primitives import padding
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

import requests as requests
import re
from bs4 import BeautifulSoup
import random
import time

from frozendict import frozendict

try:
    import ujson as json
except ImportError:
    import json as json

cache_maxsize = 64


def lru_cache_freezeargs(func):
    """
    Decorator transforms mutable dictionary and list arguments into immutable types
    Needed so lru_cache can cache method calls what has dict or list arguments.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([frozendict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: frozendict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        args = tuple([tuple(arg) if isinstance(arg, list) else arg for arg in args])
        kwargs = {k: tuple(v) if isinstance(v, list) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)

    # copy over the lru_cache extra methods to this wrapper to be able to access them
    # after this decorator has been applied
    wrapped.cache_info = func.cache_info
    wrapped.cache_clear = func.cache_clear
    return wrapped


def _extract_extra_keys_from_stores(data):
    new_keys = [k for k in data.keys() if k not in ["context", "plugins"]]
    new_keys_values = set([data[k] for k in new_keys])

    # Maybe multiple keys have same value - keep one of each
    new_keys_uniq = []
    new_keys_uniq_values = set()
    for k in new_keys:
        v = data[k]
        if not v in new_keys_uniq_values:
            new_keys_uniq.append(k)
            new_keys_uniq_values.add(v)

    return [data[k] for k in new_keys_uniq]


def decrypt_cryptojs_aes_stores(data, keys=None):
    encrypted_stores = data['context']['dispatcher']['stores']

    password = None
    if keys is not None:
        if not isinstance(keys, list):
            raise TypeError("'keys' must be list")
        candidate_passwords = keys
    else:
        candidate_passwords = []

    if "_cs" in data and "_cr" in data:
        _cs = data["_cs"]
        _cr = data["_cr"]
        _cr = b"".join(int.to_bytes(i, length=4, byteorder="big", signed=True) for i in json.loads(_cr)["words"])
        password = hashlib.pbkdf2_hmac("sha1", _cs.encode("utf8"), _cr, 1, dklen=32).hex()

    encrypted_stores = b64decode(encrypted_stores)
    assert encrypted_stores[0:8] == b"Salted__"
    salt = encrypted_stores[8:16]
    encrypted_stores = encrypted_stores[16:]

    def _EVPKDF(password, salt, keySize=32, ivSize=16, iterations=1, hashAlgorithm="md5") -> tuple:
        """OpenSSL EVP Key Derivation Function
        Args:
            password (Union[str, bytes, bytearray]): Password to generate key from.
            salt (Union[bytes, bytearray]): Salt to use.
            keySize (int, optional): Output key length in bytes. Defaults to 32.
            ivSize (int, optional): Output Initialization Vector (IV) length in bytes. Defaults to 16.
            iterations (int, optional): Number of iterations to perform. Defaults to 1.
            hashAlgorithm (str, optional): Hash algorithm to use for the KDF. Defaults to 'md5'.
        Returns:
            key, iv: Derived key and Initialization Vector (IV) bytes.

        Taken from: https://gist.github.com/rafiibrahim8/0cd0f8c46896cafef6486cb1a50a16d3
        OpenSSL original code: https://github.com/openssl/openssl/blob/master/crypto/evp/evp_key.c#L78
        """

        assert iterations > 0, "Iterations can not be less than 1."

        if isinstance(password, str):
            password = password.encode("utf-8")

        final_length = keySize + ivSize
        key_iv = b""
        block = None

        while len(key_iv) < final_length:
            hasher = hashlib.new(hashAlgorithm)
            if block:
                hasher.update(block)
            hasher.update(password)
            hasher.update(salt)
            block = hasher.digest()
            for _ in range(1, iterations):
                block = hashlib.new(hashAlgorithm, block).digest()
            key_iv += block

        key, iv = key_iv[:keySize], key_iv[keySize:final_length]
        return key, iv

    def _decrypt(encrypted_stores, password, key, iv):
        if usePycryptodome:
            cipher = AES.new(key, AES.MODE_CBC, iv=iv)
            plaintext = cipher.decrypt(encrypted_stores)
            plaintext = unpad(plaintext, 16, style="pkcs7")
        else:
            cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
            decryptor = cipher.decryptor()
            plaintext = decryptor.update(encrypted_stores) + decryptor.finalize()
            unpadder = padding.PKCS7(128).unpadder()
            plaintext = unpadder.update(plaintext) + unpadder.finalize()
            plaintext = plaintext.decode("utf-8")
        return plaintext

    if not password is None:
        try:
            key, iv = _EVPKDF(password, salt, keySize=32, ivSize=16, iterations=1, hashAlgorithm="md5")
        except:
            raise Exception("yfinance failed to decrypt Yahoo data response")
        plaintext = _decrypt(encrypted_stores, password, key, iv)
    else:
        success = False
        for i in range(len(candidate_passwords)):
            # print(f"Trying candiate pw {i+1}/{len(candidate_passwords)}")
            password = candidate_passwords[i]
            try:
                key, iv = _EVPKDF(password, salt, keySize=32, ivSize=16, iterations=1, hashAlgorithm="md5")

                plaintext = _decrypt(encrypted_stores, password, key, iv)

                success = True
                break
            except:
                pass
        if not success:
            raise Exception("yfinance failed to decrypt Yahoo data response")

    decoded_stores = json.loads(plaintext)
    return decoded_stores


_SCRAPE_URL_ = 'https://finance.yahoo.com/quote'


class TickerData:
    """
    Have one place to retrieve data from Yahoo API in order to ease caching and speed up operations
    """
    user_agent_headers = {
        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'}

    def __init__(self, ticker: str, session=None):
        self.ticker = ticker
        self._session = session or requests

    def get(self, url, user_agent_headers=None, params=None, proxy=None, timeout=30):
        proxy = self._get_proxy(proxy)
        response = self._session.get(
            url=url,
            params=params,
            proxies=proxy,
            timeout=timeout,
            headers=user_agent_headers or self.user_agent_headers)
        return response

    @lru_cache_freezeargs
    @lru_cache(maxsize=cache_maxsize)
    def cache_get(self, url, user_agent_headers=None, params=None, proxy=None, timeout=30):
        return self.get(url, user_agent_headers, params, proxy, timeout)

    def _get_proxy(self, proxy):
        # setup proxy in requests format
        if proxy is not None:
            if isinstance(proxy, dict) and "https" in proxy:
                proxy = proxy["https"]
            proxy = {"https": proxy}
        return proxy

    def get_raw_json(self, url, user_agent_headers=None, params=None, proxy=None, timeout=30):
        response = self.get(url, user_agent_headers=user_agent_headers, params=params, proxy=proxy, timeout=timeout)
        response.raise_for_status()
        return response.json()

    def _get_decryption_keys_from_yahoo_js(self, soup):
        result = None

        key_count = 4
        re_script = soup.find("script", string=re.compile("root.App.main")).text
        re_data = json.loads(re.search("root.App.main\s+=\s+(\{.*\})", re_script).group(1))
        re_data.pop("context", None)
        key_list = list(re_data.keys())
        if re_data.get("plugins"):  # 1) attempt to get last 4 keys after plugins
            ind = key_list.index("plugins")
            if len(key_list) > ind+1:
                sub_keys = key_list[ind+1:]
                if len(sub_keys) == key_count:
                    re_obj = {}
                    missing_val = False
                    for k in sub_keys:
                        if not re_data.get(k):
                            missing_val = True
                            break
                        re_obj.update({k: re_data.get(k)})
                    if not missing_val:
                        result = re_obj

        if not result is None:
            return [''.join(result.values())]

        re_keys = []    # 2) attempt scan main.js file approach to get keys
        prefix = "https://s.yimg.com/uc/finance/dd-site/js/main."
        tags = [tag['src'] for tag in soup.find_all('script') if prefix in tag.get('src', '')]
        for t in tags:
            response_js = self.cache_get(t)
            #
            if response_js.status_code != 200:
                time.sleep(random.randrange(10, 20))
                response_js.close()
            else:
                r_data = response_js.content.decode("utf8")
                re_list = [
                    x.group() for x in re.finditer(r"context.dispatcher.stores=JSON.parse((?:.*?\r?\n?)*)toString", r_data)
                ]
                for rl in re_list:
                    re_sublist = [x.group() for x in re.finditer(r"t\[\"((?:.*?\r?\n?)*)\"\]", rl)]
                    if len(re_sublist) == key_count:
                        re_keys = [sl.replace('t["', '').replace('"]', '') for sl in re_sublist]
                        break
                response_js.close()
            if len(re_keys) == key_count:
                break
        if len(re_keys) > 0:
            re_obj = {}
            missing_val = False
            for k in re_keys:
                if not re_data.get(k):
                    missing_val = True
                    break
                re_obj.update({k: re_data.get(k)})
            if not missing_val:
                return [''.join(re_obj.values())]

        return []

    @lru_cache_freezeargs
    @lru_cache(maxsize=cache_maxsize)
    def get_json_data_stores(self, sub_page: str = None, proxy=None) -> dict:
        '''
        get_json_data_stores returns a python dictionary of the data stores in yahoo finance web page.
        '''
        if sub_page:
            ticker_url = "{}/{}/{}".format(_SCRAPE_URL_, self.ticker, sub_page)
        else:
            ticker_url = "{}/{}".format(_SCRAPE_URL_, self.ticker)

        response = self.get(url=ticker_url, proxy=proxy)
        html = response.text

        # The actual json-data for stores is in a javascript assignment in the webpage
        try:
            json_str = html.split('root.App.main =')[1].split(
                '(this)')[0].split(';\n}')[0].strip()
        except IndexError:
            # Fetch failed, probably because Yahoo spam triggered
            return {}

        data = json.loads(json_str)

        # Gather decryption keys:
        soup = BeautifulSoup(response.content, "html.parser")
        keys = self._get_decryption_keys_from_yahoo_js(soup)
        # if len(keys) == 0:
        #     msg = "No decryption keys could be extracted from JS file."
        #     if "requests_cache" in str(type(response)):
        #         msg += " Try flushing your 'requests_cache', probably parsing old JS."
        #     print("WARNING: " + msg + " Falling back to backup decrypt methods.")
        if len(keys) == 0:
            keys = []
            try:
                extra_keys = _extract_extra_keys_from_stores(data)
                keys = [''.join(extra_keys[-4:])]
            except:
                pass
            #
            keys_url = "https://github.com/ranaroussi/yfinance/raw/main/yfinance/scrapers/yahoo-keys.txt"
            response_gh = self.cache_get(keys_url)
            keys += response_gh.text.splitlines()

        # Decrypt!
        stores = decrypt_cryptojs_aes_stores(data, keys)
        if stores is None:
            # Maybe Yahoo returned old format, not encrypted
            if "context" in data and "dispatcher" in data["context"]:
                stores = data['context']['dispatcher']['stores']
        if stores is None:
            raise Exception(f"{self.ticker}: Failed to extract data stores from web request")

        # return data
        new_data = json.dumps(stores).replace('{}', 'null')
        new_data = re.sub(
            r'{[\'|\"]raw[\'|\"]:(.*?),(.*?)}', r'\1', new_data)

        return json.loads(new_data)
